//   -*- C++ -*-
#ifndef _QUATERNION_H_
#define _QUATERNION_H_

// Project includes

#ifdef SGI // this can be removed when we no longer use old SGI
  #include <math.h>           // needed for : acos
#else
  #include <cmath>
#endif

#include "types.h"
#include "Lowest.h"
#include "MVector.h" // declaring class isn't enough, don't just use pointers
#include <fstream> // used for Quaternion::print
using namespace std; // for cout

class Quaternion{
  real q0, q1, q2, q3;

  public:
   // constructor
   Quaternion(real x0=1, real x1=0, real x2=0, real x3=0);

   Quaternion (MVector v, real v0=0.0);

   //this function rotates about axis by angle=abs(axis)
   void qRot(MVector axis); 

   // destructor
   ~Quaternion();

   // Quaternion product
   inline Quaternion operator*(const Quaternion& Q2) const;
   inline void       operator*=(const Quaternion& Q);
 
   // quaternian times a number
   inline Quaternion operator*(real);

   // Quaternion sum
   inline Quaternion operator+(const Quaternion& Q);
   inline void       operator+=(const Quaternion& Q);

   // Quaternion difference
   inline Quaternion operator-(const Quaternion& Q);
   inline void       operator-=(const Quaternion& Q);

   // conjugataion
   inline Quaternion operator~(void) const;

   // Quaternion divided by a number
   inline Quaternion operator/(real);

   // each component divided by a number in array real
   inline Quaternion operator/(real*);

   // Angle and the axis computed from the Quaternion
   // returned in a Quaternion.
   // Actually, these two functions  return a real and 
   // a MVector,they are packed into a Quaternion for convenience
   inline Quaternion GetAngleAndAxis(void) const ;
   inline Quaternion GetAngleAndAxis_rad(void) const ;

   // Sine of the half the angle of rotation
   inline real          GetSine(void);
   
   // computes the square of the norm
   inline real          norm(void);

   // normalizes the Quaternion

  //changed from it's previous behavior
  //
  inline void  normalize(real norm_factor);
  inline void  normalize1();

  //void normalize(void);

   // Quaternion plus a MVector
   inline Quaternion operator+(const MVector& v);

   // Quaternion times a MVector
   inline Quaternion operator*(const MVector& v);

   // assigns the MVector part to MVector v
   inline MVector getMVector();

   // assigns Quaternion elements to four variables
   inline void CopyTo(real& x0, real& x1, real& x2, real& x3);

   // computes the Quaternion that rotates the space 
   // coordinates to a coordinate system, the z axis of 
   // which is along the direction of MVector.
   inline void SetToIdent();
   inline void rotateTo(const MVector& v);
   inline void rotateToXZ(const MVector &Vx, const MVector &Vz);
  inline MVector rotate(const MVector& v) const;

   // prints the Quaternion
   inline void print(char *);
   inline void print_qua(ostream &strm);
   inline void print(ostream &fstrm, char *title); //added 06/09/00 by Xuebing

  ostream& iv_print(ostream& fout) const;

  void set_to_random_normalized();

  friend ostream& operator<< (ostream&, Quaternion) ; 
  friend istream& operator>> (istream&, Quaternion&) ; 
};



// ============================ End comment top ==========================
//
// DECLARATION OF  QUATERNION CLASS  
//

//
// Overload the operator * to perform            
// Quaternion product.                           
//
inline
Quaternion Quaternion::operator*(const Quaternion& Q2) const{
Quaternion product;

//   product.q0 = q0*Q2.q0-q1*Q2.q1-q2*Q2.q2-q3*Q2.q3;
//   product.q1 = q0*Q2.q1+q1*Q2.q0-(q2*Q2.q3-q3*Q2.q2);
//   product.q2 = q0*Q2.q2+q2*Q2.q0-(q3*Q2.q1-q1*Q2.q3);
//   product.q3 = q0*Q2.q3+q3*Q2.q0-(q1*Q2.q2-q2*Q2.q1);

 product.q0 = q0*Q2.q0-q1*Q2.q1-q2*Q2.q2-q3*Q2.q3;
 product.q1 = q0*Q2.q1+q1*Q2.q0+(q2*Q2.q3-q3*Q2.q2);
 product.q2 = q0*Q2.q2+q2*Q2.q0+(q3*Q2.q1-q1*Q2.q3);
 product.q3 = q0*Q2.q3+q3*Q2.q0+(q1*Q2.q2-q2*Q2.q1);

 return product;
}

//
// Again overload the operator * to multiply     
// each element with a constant "alpha".         
//

inline
Quaternion Quaternion::operator*(real alpha){

 return Quaternion (q0*alpha, q1*alpha, q2*alpha, q3*alpha);
}

//
// Overload the operator * to perform           
// Quaternion product with a MVector.             
//

inline
Quaternion Quaternion::operator*(const MVector& v){
Quaternion product(v);

  return *this*product;
}


//
// Overload the operator + to perform            
// Quaternion+MVector operation.                  
//

inline
Quaternion Quaternion::operator+(const MVector& v){
Quaternion sum(v);

 return *this+sum;
}

//
// sets a MVector from a Quaternion
//

// inline removed by mourad

inline
MVector Quaternion::getMVector(){

  return MVector (q1, q2, q3);
}

//
//

inline
void Quaternion::CopyTo(real& x0, real& x1, real& x2, real& x3){
  x0= q0; x1= q1; x2= q2; x3= q3;
}


//
// Overload the operator + to perform           
// Quaternion addition.                          
//

inline
Quaternion Quaternion::operator+(const Quaternion& Q){

 return Quaternion (q0+Q.q0, q1+Q.q1, q2+Q.q2, q3+Q.q3);
}


inline
Quaternion Quaternion::operator-(const Quaternion& Q){

 return Quaternion (q0-Q.q0, q1-Q.q1, q2-Q.q2, q3-Q.q3);
}

//
// Overload the operator += to perform  addition 
// and assignment.                               
//

inline
void Quaternion::operator+=(const Quaternion& Q){

 *this = *this + Q;
}


//
// Overload the operator -= to perform  addition 
// and assignment.                               
//

inline
void Quaternion::operator-=(const Quaternion& Q){

  *this = *this - Q;
}


//
// Overload the operator *= to perform  multiplication 
// and assignment.                                     
//

inline
void Quaternion::operator*=(const Quaternion& Q){

  *this = *this * Q;
}


//
// Here is the conjugate of a Quaternion.  The   
// operator ~ is overloaded.                     
//

inline
Quaternion Quaternion::operator~(void) const {

 return Quaternion(q0, -q1, -q2, -q3);
}

//
// Overload the operator / to divide each element
// of a Quaternion by the constant "alpha".      
//

inline
Quaternion Quaternion::operator/(real alpha){

 return *this * (1.0/alpha);
}

//
// Overload the operator / to divide the MVector  
// part of the Quaternion by constant numbers.   
// Used to calculate angular velocity from       
// angular momentum.                             
//

inline
Quaternion Quaternion::operator/(real *Inertia){

 return Quaternion(0, q1/Inertia[0], q2/Inertia[1], q3/Inertia[2]);
}


//
// Angle and the axis computed from the Quaternion.
//

inline
Quaternion Quaternion::GetAngleAndAxis(void) const {
  //Quaternion AngleAxis;
real Teta, SinTetaOver2;

 Teta= 2*acos(q0);
 SinTetaOver2= sin(Teta/2.0);

 if (Teta!= 0) return Quaternion(Teta*180.0/M_PI, q1/SinTetaOver2,
                                q2/SinTetaOver2, q3/SinTetaOver2);
 else return Quaternion(Teta*180.0/M_PI, 0, 0, 0);
}

inline
Quaternion Quaternion::GetAngleAndAxis_rad(void) const {
  //Quaternion AngleAxis;
real Teta, SinTetaOver2;

 if (q0>0) { 
   Teta= 2*acos(q0);
   SinTetaOver2 = sqrt(1.0-q0*q0);
 }
 else {
   Teta = 2*acos(-q0);           //to make sure that 0<Teta<PI
   SinTetaOver2 = -sqrt(1.0-q0*q0);  //the result should be the same
 }                                  //for two equivalent quaternions

 if (Teta!=0 ) return Quaternion(Teta, q1/SinTetaOver2,
                                q2/SinTetaOver2, q3/SinTetaOver2);
 else return Quaternion(0, 0, 0, 0);
}


//
// Given a MVector in space coordinates, this routine returns the  
// Quaternion that rotates the space coordinates to a coordinate  
// system, the z axis of which is along the direction of MVector   
// r, sent to the routine as the input.                           
// The coordinate system is rotated by an angle acos(r.z/(r*r))   
// around an axis perpendicular to (r.x, r.y, 0).                 
//

inline
void Quaternion::SetToIdent() {q0=1;q1=0;q2=0;q3=0;}

inline
void Quaternion::rotateTo(const MVector& l){

// changed by mourad from "real phi" to "S phi"
  //cout<<"rotateTo, l="<<l<<endl;
  if(l.IsZero()) {SetToIdent(); return;}
  real phi= l.azimuthal();

  // The MVector n around which we rotate.
  MVector n(cos(phi+M_PI/2), sin(phi+M_PI/2), 0.0);

// idem changed by mourad
   
  real teta= l.theta();
// idem changed by mourad
   
  real SinTetaOverTwo= sin(teta/2.0);

  *this= ~Quaternion(-n*SinTetaOverTwo, cos(teta/2.0)); 
}

//added by DMA
//create quaternion that rotates x and z axes to Vx and Vz
//used to find the quaternion from M in cell:update_m

inline
void Quaternion::rotateToXZ(const MVector &Vx, const MVector &Vz)
{
  assert (Vx*Vz<(0.001*Vx.Magnitude()*Vz.Magnitude()));
  Quaternion Q1, Q2;
  MVector x(1,0,0), xt, axis;
  real angle;

  Q1.rotateTo(Vz);
  xt = Q1.rotate(x);
  angle = acos(xt*Vx/(xt.Magnitude()*Vx.Magnitude()));
  axis = (xt^Vx).normalize()*angle;
  Q2.qRot(axis);
  *this = (Q2 * Q1);
}

inline
MVector Quaternion::rotate(const MVector& V) const {
//    PRINT3(*this);
//    PRINT3(V);
//    PRINT3((*this)*V);
//    PRINT3((~(*this)));
//    PRINT3((*this)*V*(~(*this)));
  return ((*this)*V*(~(*this))).getMVector();
}


//
// a Quaternion is printed here.                
//

inline
void Quaternion::print(char *title){

  cout << title <<"=(" << q0;
  cout << ", " << q1;
  cout << ", " << q2;
  cout << ", " << q3 << ")"<<endl;
}

//
//
//

inline
void Quaternion::print_qua(ostream &strm){
  strm<<" "<< q0 << "  " << q1 << "  " << q2 << "  " << q3 << " ";
}

inline
void Quaternion::print(ostream &fstrm, char *title){

  fstrm<< title <<"=(" << q0;
  fstrm<< ", " << q1;
  fstrm<< ", " << q2;
  fstrm<< ", " << q3 << ")"<<endl;

}

//
// Square of the norm is calculated.             
//

inline
real Quaternion::norm(void){

 return q0*q0+q1*q1+q2*q2+q3*q3;
}



  // What's going here : this is the bahavior implemented by Gunal!

//
//Quaternion is normalized to unity.            
//

inline
void Quaternion::normalize(real normfactor){
 normfactor = (3-normfactor)/2.0;

 //normfactor= 1.0/sqrt(normfactor);

 *this = *this * normfactor;

 cout <<"qnorm\t"<<norm()<<endl;

}

inline
void Quaternion::normalize1(){
  *this = *this / sqrt(this->norm());
}



// //
// // Quaternion is normalized to unity.            
// //
// 
// void Quaternion::normalize()
// {
  
//   //  S temp=q0*q0+q1*q1+q2*q2+q3*q3;
//   //  temp=(3-temp)/2.0;
//   //  *this = *this * temp;
  
//   *this = *this / sqrt(q0*q0+q1*q1+q2*q2+q3*q3);
// }

//
// Sine of the half the angle of rotation.       
//

inline
real Quaternion::GetSine(){

 return sqrt(q1*q1+q2*q2+q3*q3);
}


// ========================== Start comment tail =========================
#endif  // _QUATERNION_H_
