//mpole.C file

#include "FMM.h"
#include "mpole.h"
#include "System.h"  //DMAfs

//  extern System *My;  //DMAfs

//overloaded operator to output mpole
ostream &operator<<(ostream &out, const mpole &right)
{
  out<<" maximum order is "<<right.order<<"\nvalues:\n";
  for (int orderTemp = 0; orderTemp <= right.order; orderTemp++)
  {
    out<<"  order "<<orderTemp<<" ( ";
  //start from i=0 if orderTemp==0
    for (int i = (orderTemp>0 ? My->MyFMM->NumBelow[orderTemp-1] : 0); 
      i<My->MyFMM->NumBelow[orderTemp]; i++)
      out<<right.values[i]<<"   ";
    out<<") \n";
  }
  return out;
}


//overloaded operator to input mpole
istream& operator>>(istream& input, mpole& right)
{
  char junk[20]; 
  int orderTemp;
  input>>junk>>junk>>junk;
  assert ((junk[0] == 'i') && (junk[1] == 's'));  //must be "is"
  input>>right.order;
  input.getline (junk, '\n');
  input.getline (junk, '\n');
  assert (junk[0] == 'v');  //must be "values"

  for (int orderCur = 0; orderCur <= right.order; orderCur++)
  {
    input>>junk>>orderTemp;
    assert (orderTemp == orderCur);
    assert (junk[0] == 'o'); //must be "order"
    input>>junk;

    for (int i = (orderTemp>0 ? My->MyFMM->NumBelow[orderTemp-1] : 0); 
        i<My->MyFMM->NumBelow[orderTemp]; i++)
      input>>right.values[i];
    input.getline (junk, '\n'); 
  }
  return input;
}

mpole::mpole(int orderIn)//default constructor
{
  cout<<"creading mpole, order="<<order<<endl;
  if(orderIn<0) // flag for use max
	  order = My->MyFMM->MaxOrder; // default value
  else order = orderIn;    
  values = new real[My->MyFMM->NumBelow[order]];
  assert(values);
  for (int i = 0; i < My->MyFMM->NumBelow[order]; i++)//initialize all values to zero
    values[i] = 0.0;
}

mpole::mpole (const mpole &right)
{
  order = right.order;
  values = new real[My->MyFMM->NumBelow[order]];
  assert (values);

  for (int i=0; i < My->MyFMM->NumBelow[order]; i++)
    values[i] = right.values[i];
}

mpole::~mpole()//destructor
{
  delete [] values;//need to deallocate memory
  values = NULL;
}


real& mpole::GetValue(int i, int j, int k)   //returns value[i][j][k]
{
  assert ((i+j+k)<=order);
  return values[My->MyFMM->index[i][j][k]];
}


const real& mpole::GetValue(int i, int j, int k) const  //constant version
{
  assert ((i+j+k)<=order);
  return values[My->MyFMM->index[i][j][k]];
}


void mpole::SetToZero()//set all values to zero
{
  for (int i = 0; i<My->MyFMM->NumBelow[order]; i++)
    values[i] = 0.0;
}


real mpole::compare(const mpole *right)//compare mpole right with this
{
  real error = 0.0, errorCurrent;
  assert (order==right->order);  //we can compare mpoles with the same order
  for (int i = 0; i<My->MyFMM->NumBelow[order]; i++)
  {
    errorCurrent = fabs(values[i] - right->values[i]);
    if (errorCurrent > error)
      error = errorCurrent;
  }
  return error;  //returns the maximum difference between the two mpoles
}


void mpole::rescale (real factor) //divide values by factor^(n+1)
{
  assert (factor!=0.);
  int curOrder=0;
  real term = 1./factor;

  for (int i=0; i < My->MyFMM->NumBelow [order]; i++)
  {
    if (My->MyFMM->IndArr[i].sum() > curOrder)    //order is increased
    {
      term/=factor;
      curOrder++;         
    }
    values[i]*=term;
  }
  assert (curOrder == order);
}

void mpole::SetToVector(const MVector &c)
//regard vector as a special case of mpole
{
  SetToZero();
  values[1] = c[0];
  values[2] = c[1];
  values[3] = c[2];
}


void mpole::SetToSimplePowers(const MVector &c)
//this = cx^nx*cy^ny*cz^nz
{
  SetToZero();
  values[0]=1.;
  values[1]=c[0];
  values[2]=c[1];
  values[3]=c[2];

  for (int i=4; i < My->MyFMM->NumBelow[order]; i++)
    values[i]=values[My->MyFMM->RecurseMpole[i]]*c[My->MyFMM->RecurseAxis[i]];
}

void mpole::SetToPowers(const MVector &c)
{  // this[n] = cx^nx*cy^ny*cz^nz / nx!ny!nz!
  SetToZero();
  values[0]=1.;
  values[1]=c[0];
  values[2]=c[1];
  values[3]=c[2];

  for (int i=4; i < My->MyFMM->NumBelow[order]; i++)
    values[i]=values[My->MyFMM->RecurseMpole[i]]*c[My->MyFMM->RecurseAxis[i]]
             *My->MyFMM->RecurseFactor[i];
}

void mpole::SetToDeriv(const MVector &c)
{   //this = D_n(r) in equation for derivatives
  SetToZero();
  My->MyFMM->scratchMpole2->SetToSimplePowers(c);
  assert(c.Square());//to check that c in non-zero
  real c2 = 1.0/c.Square();
  real factor = sqrt(c2);
  int iEnd = 0, iBegin = 0;

  for (int ord1 = 0; ord1 <=order; ord1++)//loop order
  {//such initialization is needed for speeding up the performance
    iBegin = iEnd; iEnd = My->MyFMM->NumBelow[ord1];

    for (int i = iBegin; i < iEnd; i++)
    {
      real D = 0.0;
      for (int j = 0; j < My->MyFMM->DerivTerms[i]; j++)
        D += (My->MyFMM->DerivCoef[i][j])*My->MyFMM->scratchMpole2->values[My->MyFMM->DerivPower[i][j]];

      values[i] = D*factor;
    }
    factor *= c2;
  }
}


//used for periodic FMM (DMApf)
void mpole::SetToPowerSym (const int* periodicity)
{
//  cout<<"\n\nStarting Sym"<<endl;
  mpole *tempMpole = new mpole(-1);
  int3 t;
  SetToZero ();

  //sum over all t's:
  for (t.values [0]=(periodicity[0] ? -1 : 0);
       t.values [0]<= (periodicity[0] ? 1 :0); t.values[0]++)
    for (t.values [1]=(periodicity[1] ? -1 : 0);
         t.values [1]<= (periodicity[1] ? 1 :0); t.values[1]++)
      for (t.values [2]=(periodicity[2] ? -1 : 0);
           t.values [2]<= (periodicity[2] ? 1 :0); t.values[2]++)

      {
        tempMpole->SetToPowers (MVector (t[0]*My->MyFMM->reducedSize[0],
          t[1]*My->MyFMM->reducedSize[1], t[2]*My->MyFMM->reducedSize[2])*(-.5));
        *this += (*tempMpole)*My->MyFMM->calculateSymFactor(t);
      }
  delete tempMpole;
}


void mpole::AddTaylor(const mpole *Q, const MVector &c)
{
  invert(); //inefficient: double inverting, SEE LATER
  My->MyFMM->scratchMpole1->SetToDeriv(c);
  AddDownConv(My->MyFMM->scratchMpole1, Q);
  invert();//to invert only the down-convolution of D and Q
}

void mpole::AddUpConv(const mpole *m1, const mpole *m2)
{ // add up-convolution of m1 and m2 (see Eq. ... of paper) to this mpole.
  int orderMin = m1->order;
  if (orderMin>order) orderMin = order;
  for (int im = 0; im < My->MyFMM->NumBelow[m2->order]; im++) //im was changed to start from 1
  {                                                       // (we don't have monopoles)
    int m = (My->MyFMM->IndArr[im]).sum();

    for (int ip = 0; ip < My->MyFMM->NumBelow[orderMin - m]; ip++)
      values[My->MyFMM->ConvSum[im][ip]]+=m1->values[ip]*m2->values[im];
  }//end for im-loop
}

void mpole::AddDownConv(const mpole *m1, const mpole *m2)
{
  int orderMin = m1->order;
  if (orderMin>m2->order) orderMin = m2->order;
  for (int im = 0; im < My->MyFMM->NumBelow[order]; im++)  //im was changed to start from 1
  {                                                    // (we don't have monopoles)
    int m = (My->MyFMM->IndArr[im]).sum();

    for (int ip = 0; ip < My->MyFMM->NumBelow[orderMin - m]; ip++)
      values[im] += m1->values[My->MyFMM->ConvSum[im][ip]]*m2->values[ip];
  }//end for im-loop
}//end of AddDownConv

void mpole::invert() // reverses sign of odd-order components
{
  for (int i = 0; i<My->MyFMM->NumOddBelow[order]; i++)
    values[My->MyFMM->OddOnes[i]]*=-1;
}

void mpole::AddShift(const mpole *Q, const MVector &c)
{
  My->MyFMM->scratchMpole1->SetToPowers(c);
  AddUpConv(My->MyFMM->scratchMpole1, Q);
}

void mpole::print(char * title, ofstream &out)//output an array
{
  //we want to restore the initial settings
  long originalPrecision = out.precision(); 

  out.precision(4);
  out<<title<<" is:";
// zero order:
  out<<"\n\torder = 0: "<<values[0]<<endl;
//first order:
  out<<"\n\torder = 1:  "<<values[1]<<"  "<<values[2]<<"  "<<values[3]<<endl;
//second order: output as a matrix:  
  out<<"\n\torder = 2:  "<<values[4]<<"  "<<values[5]<<"  "<<values[7]
     <<"\n\t            "<<values[5]<<"  "<<values[6]<<"  "<<values[8]
     <<"\n\t            "<<values[7]<<"  "<<values[8]<<"  "<<values[9]<<endl;
  for (int orderTemp = 3; orderTemp <= order; orderTemp++)
  {
    out<<"\n\torder = "<<orderTemp<<":\n";
    for (int i = My->MyFMM->NumBelow[orderTemp-1]; i < My->MyFMM->NumBelow[orderTemp]; i++)
      out<<"\t\tn = "<<My->MyFMM->IndArr[i][0]<<" "<<My->MyFMM->IndArr[i][1]<<" "
        <<My->MyFMM->IndArr[i][2]<<"\t"<<values[i]<<endl;
  }
  out.precision(originalPrecision);
}


//dot operator for mpoles
real mpole::operator* (const mpole& right) const
{
  int smallestOrder = order;
  if (smallestOrder > right.order)
    smallestOrder = right.order;

  real dotProduct = 0.0;

  for (int i=0; i<My->MyFMM->NumBelow[smallestOrder]; i++)
    dotProduct += values[i]*right.values[i];

  return dotProduct;
}


//multiplication by a real number:
mpole operator* (const mpole& m, const real &r)
{
  mpole result (m.order);

  for (int i=0; i < My->MyFMM->NumBelow [result.order]; i++)
    result.values[i] = m.values[i] * r;

  return result;
}

//division by a real number:
mpole mpole::operator/ (const real &r) const
{
  mpole result (order);

  for (int i=0; i < My->MyFMM->NumBelow [order]; i++)
    result.values[i] = values[i] / r;

  return result;
}

mpole mpole::operator+ (const mpole& m) const
{
  assert (order == m.order);
  mpole result (order);

  for (int i=0; i<My->MyFMM->NumBelow[order]; i++)
    result.values[i] = values[i] + m.values[i];

  return result;
}

const mpole& mpole::operator+= (const mpole &m)
{
  assert (order==m.order);

  for (int i=0; i<My->MyFMM->NumBelow[order]; i++)
    values[i] += m.values[i];

  return *this;
}

const mpole& mpole::operator*= (const real& r)
{
  for (int i=0; i<My->MyFMM->NumBelow[order]; i++)
    values[i] *= r;

  return *this;
}

const mpole& mpole::operator= (const mpole& m)
{
  assert (m.order==order);  //to catch errors

  for (int i=0; i<My->MyFMM->NumBelow[order]; i++)
    values[i] = m.values[i];

  return *this;
}
